# Standard Library
from typing import Any

# Third Party
from ninja import NinjaAPI

# Django
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Memberaudit Doctrine Checker
from madc import __title__, providers
from madc.api import schema
from madc.api.helpers import (
    generate_button,
    get_alts_queryset,
    get_main_character,
    get_manage_permission,
)
from madc.models import SkillList

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class DoctrineCheckerApiEndpoints:
    tags = ["Doctrine Checker"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "{character_id}/doctrines/",
            response={200: list[schema.CharacterDoctrines], 403: str},
            tags=self.tags,
        )
        def get_doctrines(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)
            
            # Get the skill lists for the main character
            skilllists = providers.skills.get_user_skill_list(
                user_id=main.character_ownership.user_id
            )
            
            # Active skill lists are the ones that are visible in the UI
            visibles = list(
                SkillList.objects.filter(active=1).values_list("name", flat=True)
            )
            
            output = {}

            for c in characters:
                output[c.character_id] = {
                    "character": c,
                    "doctrines": {},
                    "skills": {},
                }

            for k, s in skilllists["skills_list"].items():
                for k, d in s["doctrines"].items():
                    # filter out hidden items
                    if k in visibles:
                        output[s["character_id"]]["doctrines"][k] = d
                # Add skills to the character
                output[s['character_id']]["skills"] = s["skills"]

            return list(output.values())

        @api.get(
            "administration/",
            response={200: Any, 403: str},
            tags=self.tags,
        )
        def admin_doctrines(request):
            character_id = request.user.profile.main_character.character_id
            response, __ = get_manage_permission(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            skilllist_obj = SkillList.objects.all().order_by("ordering", "name")

            skilllist_dict = {}

            btn_template = "madc/partials/form/button.html"
            url = reverse(
                viewname="madc:delete_doctrine",
            )

            settings_dict = {
                "title": _("Delete Skill Plan"),
                "color": "danger",
                "icon": "fa fa-trash",
                "text": _("Are you sure you want to delete this skill plan?"),
                "modal": "skillplan-delete",
                "action": url,
                "ajax": "action",
            }

            for skill_list in skilllist_obj:
                edit_btn = generate_button(
                    pk=skill_list.pk,
                    template=btn_template,
                    queryset=skilllist_obj,
                    settings=settings_dict,
                    request=request,
                )

                skilllist_dict[skill_list.name] = {
                    "name": skill_list.name,
                    "skills": skill_list.get_skills(),
                    "active": skill_list.active,
                    "ordering": skill_list.ordering,
                    "actions": {
                        "delete": format_html(edit_btn),
                    },
                }

            return skilllist_dict



# Generated by Django 4.2.20 on 2025-07-03 10:22

import django.core.validators
from django.db import migrations, models
import madc.models.skillchecker


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='General',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('basic_access', 'Can access this app, Voices of War.'), ('manage_access', 'Has access to moderation tools'), ('admin_access', 'Has access to all Administration tools')),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='SkillList',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(default=None, max_length=500, null=True)),
                ('skill_list', models.TextField(default='', null=True, validators=[madc.models.skillchecker.skillvalidator])),
                ('ordering', models.IntegerField(default=0, validators=[django.core.validators.MinValueValidator(0)], verbose_name='Order Weight')),
                ('active', models.BooleanField(default=True, help_text='If unchecked, this skill list will not be used in the skill checker.', verbose_name='Active')),
                ('category', models.CharField(default=None, help_text='Category of the skill list, used for grouping.', max_length=100, null=True, verbose_name='Category')),
            ],
        ),
    ]



# Django
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.safestring import mark_safe

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Memberaudit Doctrine Checker
from madc import __title__, models

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_manage_permission(request, character_id):
    """Check if the user has permission to manage."""
    perms = True
    main_char = EveCharacter.objects.select_related(
        "character_ownership",
        "character_ownership__user__profile",
        "character_ownership__user__profile__main_character",
    ).get(character_id=character_id)
    try:
        main_char = main_char.character_ownership.user.profile.main_character
    except ObjectDoesNotExist:
        pass

    # check access
    visible = models.SkillList.objects.manage_to(request.user)
    if main_char not in visible:
        account_chars = (
            request.user.profile.main_character.character_ownership.user.character_ownerships.all()
        )
        if main_char in account_chars:
            pass
        else:
            perms = False
    return perms, main_char


def get_main_character(request, character_id):
    perms = True
    main_char = EveCharacter.objects.select_related(
        "character_ownership",
        "character_ownership__user__profile",
        "character_ownership__user__profile__main_character",
    ).get(character_id=character_id)
    try:
        main_char = main_char.character_ownership.user.profile.main_character
    except ObjectDoesNotExist:
        pass

    # check access
    visible = models.SkillList.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        account_chars = (
            request.user.profile.main_character.character_ownership.user.character_ownerships.all()
        )
        if main_char in account_chars:
            pass
        else:
            perms = False
    return perms, main_char


def get_alts_queryset(main_char):
    try:
        linked_characters = (
            main_char.character_ownership.user.character_ownerships.all().values_list(
                "character_id", flat=True
            )
        )

        return EveCharacter.objects.filter(id__in=linked_characters)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def generate_button(pk: int, template, queryset, settings, request) -> mark_safe:
    """Generate a html button for the tax system"""
    return format_html(
        render_to_string(
            template,
            {
                "pk": pk,
                "queryset": queryset,
                "settings": settings,
            },
            request=request,
        )
    )


# Django
from django.db import models

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger


from app_utils.logging import LoggerAddTag
# AA Memberaudit Doctrine Checker
from madc import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

class SkillListQuerySet(models.QuerySet):
    def manage_to(self, user):
        qs = EveCharacter.objects.get_queryset()
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all Data for superuser %s.", user)
            return qs.all()

        if user.has_perm("voicesofwar.admin_access"):
            logger.debug("Returning all Data for %s.", user)
            return qs.all()

        logger.debug("User %s has no permission. Nothing visible.", user)
        return qs.none()


class SkillListManagerBase(models.Manager):
    def get_queryset(self):
        return SkillListQuerySet(self.model, using=self._db)

    @staticmethod
    def visible_eve_characters(user):
        qs = EveCharacter.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return qs.all()

        if user.has_perm("voicesofwar.admin_access"):
            logger.debug("Returning all characters for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            assert char
            # build all accepted queries
            queries = [models.Q(character_ownership__user=user)]

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )
            # filter based on queries
            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()

    def manage_to(self, user):
        return self.get_queryset().manage_to(user)


SkillListManager = SkillListManagerBase.from_queryset(SkillListQuerySet)


# Standard Library
import json
from hashlib import md5

# Django
from django.contrib.auth.models import User
from django.core.cache import cache
from django.utils.html import format_html

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from memberaudit.models import CharacterSkill

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.models.skillchecker import SkillList

SKILL_CACHE_TIMEOUT_SECONDS = 60 * 60 * 72  # 72 hours
SKILL_CACHE_HEADERS_KEY = "SKILL_HEADER"
SKILL_CACHE_USER_KEY = "SKILL_LISTS_{}"

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class SkillListHandler:
    """handler for skill lists."""

    def _get_skill_list_hash(self, skills: list[str]) -> str:
        """Generate a hash for the skill list names."""
        return md5(",".join(str(x) for x in sorted(skills)).encode()).hexdigest()

    def _get_chars_hash(self, characters: list[int]) -> str:
        return md5(",".join(str(x) for x in sorted(characters)).encode()).hexdigest()

    def _build_account_cache_key(self, characters):
        """Build a cache key based on the character IDs."""
        return SKILL_CACHE_USER_KEY.format(self._get_chars_hash(characters))

    def check_skill_lists(self, skill_lists: list[SkillList], linked_characters):
        """Check if the skill lists are up to date."""

        skills = (
            CharacterSkill.objects.filter(
                character__eve_character__character_id__in=linked_characters
            )
            .select_related(
                "eve_type", "eve_type__eve_group", "character__eve_character"
            )
            .order_by("eve_type__name")
        )

        skill_dict = {}
        skill_list_dict = {}

        for skill in skills:
            character = skill.character.eve_character.character_name
            skill_group = skill.eve_type.eve_group.name

            # Create a new entry for the character if it doesn't exist
            if character not in skill_dict:
                skill_dict[character] = {
                    "character_id": skill.character.eve_character.character_id,
                    "skills": {},
                }

            # Add the skill to the dictionary
            skill_dict[character]["skills"][skill.eve_type.name] = {
                "skill_group": skill_group,
                "sp_total": skill.skillpoints_in_skill,
                "active_level": skill.active_skill_level,
                "trained_level": skill.trained_skill_level,
            }

        # Get the skill lists and their skills
        for skill_list in skill_lists:
            skill_list_dict[skill_list.name] = skill_list.get_skills()

        for character, character_data in skill_dict.items():
            character_data["doctrines"] = {}
            for skill_list_name, skills in skill_list_dict.items():
                character_data["doctrines"][skill_list_name] = {}
                for skill, level in skills.items():
                    level = int(level)
                    if level > character_data["skills"].get(skill, {}).get(
                        "active_level", 0
                    ):
                        character_data["doctrines"][skill_list_name][skill] = level

                # TODO: Make a Helper to Handle HTML Formatting
                # Add Modal Overview for missing skills for each Character
                # Add HTML to each individual doctrine
                if character_data["doctrines"][skill_list_name]:
                    character_data["doctrines"][skill_list_name]["html"] = format_html(
                        """
                            <div role="group" class="btn-group">
                                <button type="button" class="btn btn-danger btn-sm" id="missing-{}-{}">
                                    {}
                                </button>
                                <button type="button" class="flex-one btn btn-danger btn-sm" id="copy-{}-{}">
                                    <i class="fa-solid fa-copy"></i>
                                </button>
                            </div>
                        """,
                        skill_list_name,
                        character_data["character_id"],
                        skill_list_name,
                        character_data["character_id"],
                        md5(skill_list_name.encode()).hexdigest(),
                    )
                else:
                    character_data["doctrines"][skill_list_name]["html"] = format_html(
                        """
                            <div role="group" class="btn-group">
                                <button type="button" class="btn btn-success btn-sm">
                                    {}
                                </button>
                                <button type="button" class="flex-one btn btn-success btn-sm">
                                    <i class="fa-solid fa-check"></i>
                                </button>
                            </div>
                        """,
                        skill_list_name,
                    )
        return skill_dict

    def get_user_skill_list(self, user_id: int, force_rebuild=False) -> dict:
        """Get the skill list for a user."""
        linked_characters = (
            User.objects.get(id=user_id)
            .character_ownerships.all()
            .values_list("character__character_id", flat=True)
        )
        skill_lists = SkillList.objects.all().order_by("ordering", "name")
        skill_list_hash = self._get_skill_list_hash(skill_lists.values_list("name"))
        account_key = self._build_account_cache_key(linked_characters)
        cached_header = cache.get(SKILL_CACHE_HEADERS_KEY, False)
        skill_lists_up_to_date = cached_header == skill_list_hash

        if skill_lists_up_to_date and not force_rebuild:
            cached_skills = cache.get(account_key, False)

            if cached_skills is not False:
                logger.debug(
                    "Using cached skill list for user %s: %s",
                    user_id,
                    account_key,
                )
                cached_skills = json.loads(cached_skills)
                if cached_skills.get("doctrines", False) == skill_list_hash:
                    return cached_skills

        output = {
            "doctrines": skill_list_hash,
            "characters": account_key,
            "skills_list": self.check_skill_lists(skill_lists, linked_characters),
        }

        output_json = json.dumps(output)

        cache.set(
            account_key,
            output_json,
            SKILL_CACHE_TIMEOUT_SECONDS,
        )
        cache.set(SKILL_CACHE_HEADERS_KEY, skill_list_hash)
        return output
